"Status messages for querying Actors or ActorSystems"

from thespian.actors import ActorSystemMessage


# ----------------------------------------------------------------------
# ActorSystem diagnostics
#
# An ActorSystem should support the following messages which allow
# diagnostic queries and monitoring.

class Thespian_StatusReq(ActorSystemMessage):
    "Sent to an ActorSystem or an Actor to request a status response"
    pass


class _Common_StatusResp(ActorSystemMessage):
    def __init__(self):
        self.pendingMessages  = []  # array of (fromActor, toActor, msgstr) tuples
        self.pendingWakeups   = {}  # key = wakeup time, value = array of WakeupMessages
        self.receivedMessages = []  # array of (fromActor, toActor, msgstr) tuples
        self.childActors      = []  # array of addresses
        self._numSent         = 0
        self._numSendFailures = 0
        self._numReceived     = 0
        self.miscKeyVals      = {}
    def addChild(self, childAddress): self.childActors.append(childAddress)
    def addPendingMessage(self, fromActor, toActor, msgstr):
        self.pendingMessages.append( (fromActor, toActor, msgstr) )
    def addReceivedMessage(self, fromActor, toActor, msgstr):
        self.receivedMessages.append( (fromActor, toActor, msgstr) )
    def addWakeups(self, wakeupdict):
        self.pendingWakeups.update(wakeupdict)
    def addSent(self, count): self._numSent += count
    def addSendFailures(self, count): self._numSendFailures += count
    def addReceived(self, count): self._numReceived += count
    def addKeyVal(self, key, value): self.miscKeyVals[key] = value


class Thespian_SystemStatus(_Common_StatusResp):
    "ActorSystem response to a Thespian_StatusReq"
    def __init__(self, address, conventionLeader=None, capabilities=None, inShutdown=False):
        super(Thespian_SystemStatus, self).__init__()
        self.adminAddress        = address
        self.conventionLeader    = conventionLeader
        self.capabilities        = {} if capabilities is None else capabilities
        self.conventionRegisterTime = None
        self.conventionAttendees = []  # array of (Admin actorAddress, last checkin or remaining time)
        self.deadLetterHandler   = None
        self.deadLetterAddresses = []
        self.globalActors = {}
        self.inShutdown  = inShutdown
    def addDeadLetter(self, deadAddress): self.deadLetterAddresses.append(deadAddress)
    def setConventionLeaderAddress(self, addr): self.conventionLeader = addr
    def addConventioneer(self, memberAddress, validTime):
        self.conventionAttendees.append( (memberAddress, validTime) )
    def setConventionRegisterTime(self, time): self.conventionRegisterTime = time
    def addGlobalActor(self, name, address): self.globalActors[name] = address
    def setDeadLetterHandler(self, address): self._deadLetterHandler = address


# Likewise the management of each Actor should support the following diagnostics

class Thespian_ActorStatus(_Common_StatusResp):
    def __init__(self, address, actorClass, adminAddress, parentAddress=None):
        super(Thespian_ActorStatus, self).__init__()
        self.actorAddress    = address
        self.actorClass      = actorClass
        self.adminAddress    = adminAddress
        self.parentAddress   = parentAddress



def _common_formatStatus(response, childActorTag, showAddress=str):
    fmtResp = []
    fmtResp.append('  |%s Actors [%d]:'%(childActorTag, len(response.childActors)))
    fmtResp.extend(['    @ %s'%(showAddress(A)) for A in response.childActors])
    fmtResp.append('  |Pending Messages [%d]:'%len(response.pendingMessages))
    fmtResp.extend(['    %s --> %s:  %s'%(showAddress(F), showAddress(T), M)
                    for F,T,M in response.pendingMessages])
    fmtResp.append('  |Received Messages [%d]:'%len(response.receivedMessages))
    fmtResp.extend(['    %s <-- %s:  %s'%(showAddress(T), showAddress(F), M)
                    for F,T,M in response.receivedMessages])
    fmtResp.append('  |Pending Wakeups [%d]:'%len(response.pendingWakeups))
    import datetime
    fmtResp.extend(['    %s%s'%(str(W),
                   (' (in %s)'%(W - datetime.datetime.now()) if isinstance(W, datetime.datetime)
                    else '')) for W in response.pendingWakeups])
    if response.miscKeyVals:
        miscKeys = list(response.miscKeyVals.keys())
        miscKeys.sort()
        maxlen = max([len(K) for K in miscKeys])
        fmtResp.extend(['  |> %%%ds: %%%%s' % maxlen % K % str(response.miscKeyVals[K])
                        for K in miscKeys])
    return fmtResp


def formatStatus(response, showAddress=str):
    fmtResp = []
    if isinstance(response, Thespian_SystemStatus):
        fmtResp.append('Status of ActorSystem @ %s:%s'%(showAddress(response.adminAddress),
                                                        '  [IN SHUTDOWN]' if response.inShutdown else ''))
        fmtResp.append('  |Capabilities[%d]:'%len(response.capabilities))
        fmtResp.extend(['    %29s: %s'%(k, response.capabilities[k]) for k in response.capabilities])
        if response.conventionLeader:
            fmtResp.append('  |Convention Leader: %s'%(showAddress(response.conventionLeader)))
            if response.conventionRegisterTime:
                fmtResp.append('  Registration valid %s'%(str(response.conventionRegisterTime)))
        elif response.conventionAttendees:
            fmtResp.append('  Appears to be the Convention Leader')
        fmtResp.append('  |Convention Attendees [%d]:'%len(response.conventionAttendees))
        fmtResp.extend(['    @ %s: %s'%(showAddress(addr), str(validtime))
                        for addr,validtime in response.conventionAttendees])
        fmtResp.extend(_common_formatStatus(response, 'Primary', showAddress))
        if response.deadLetterHandler:
            fmtResp.append('  |DeadLetter Handler: %s'%showAddress(response.deadLetterHandler))
        fmtResp.append('  |DeadLetter Addresses [%d]:'%len(response.deadLetterAddresses))
        fmtResp.extend(['    %s'%(showAddress(A)) for A in response.deadLetterAddresses])
    elif isinstance(response, Thespian_ActorStatus):
        fmtResp.append('Status of %s Actor @ %s:'%(response.actorClass,
                                                   showAddress(response.actorAddress)))
        fmtResp.append('  |Administrator: %s'%(showAddress(response.adminAddress)))
        fmtResp.append('  |Parent  Actor: %s'%(showAddress(response.parentAddress)))
        fmtResp.extend(_common_formatStatus(response, 'Child', showAddress))
    else:
        fmtResp.append('Status Query Response: %s'%(str(response)))
    return fmtResp
