from cStringIO import StringIO
from zipfile import ZipFile
from os import path as ospath
import logging
import imp
import sys
from thespian.actors import InvalidActorSourceHash


if sys.version_info >= (3,0):
    exec("def do_exec(co, loc): exec(co, loc)\n")
else:
    exec("def do_exec(co, loc): exec co in loc\n")


class HashLoader(object):
    def __init__(self, finder, isModuleDir=False):
        self.finder = finder
        self.isModuleDir = isModuleDir
    def load_module(self, moduleName):
        if moduleName in sys.modules:
            return sys.modules[moduleName]
        mod = sys.modules.setdefault(moduleName, imp.new_module(moduleName))
        mod.__file__ = moduleName
        mod.__loader__ = self
        if self.isModuleDir:
            mod.__path__ = moduleName
        mod.__package__ = moduleName.rpartition('.')[0]
        name = ospath.join(*tuple(moduleName.split('.')[1:])) + ('/__init__' if self.isModuleDir else '') + '.py'
        try:
            code = self.finder.withZipElementSource(name,
                                                    lambda s: compile(s, mod.__file__, "exec"))
            do_exec(code, mod.__dict__)
        except Exception as ex:
            print('loadfail: %s'%str(ex))
            del sys.modules[moduleName]
            return None
        return mod

class HashRootLoader(object):
    """The SourceHashFinder below inserts the hashRoot at the beginning of the
       import path to ensure that the sources imported from the
       hashedSource are in a separate namespqce.  This HashRootLoader
       object "eats" that top-level hashRoot namespace from the
       beginning of import paths.
    """
    def __init__(self, finder):
        self.finder = finder
    def load_module(self, moduleName):
        if moduleName != self.finder.hashRoot(): return None
        mod             = sys.modules.setdefault(moduleName, imp.new_module(moduleName))
        mod.__file__    = moduleName
        mod.__loader__  = HashLoader(self.finder)
        mod.__path__    = moduleName
        mod.__package__ = moduleName
        code = compile('', mod.__file__, "exec")
        do_exec(code, mod.__dict__)
        return mod


class SourceHashFinder(object):
    """This module finder looks in the specified hashedSource for the
       indicated module to import and returns an appropriate HashLoader object if
       the module is in that hashedSource.
    """
    def __init__(self, srcHash, decryptor, enczfsrc):
        self.decryptor = decryptor
        self.enczfsrc = enczfsrc
        self.srcHash = srcHash
    def hashRoot(self):
        # All imports that come from a hashedSource will have the
        # hashRoot automatically inserted as the start of the import
        # namespace.  This helps to keep imports from different
        # hashedSource from conflicting or polluting the caller's
        # namespace (which is also a good security measure against
        # malware injection).
        return '{{' + self.srcHash + '}}'
    def _getFromZipFile(self, getter):
        plainsrc = self.decryptor(self.enczfsrc)
        z = ZipFile(StringIO(plainsrc))
        try:
            return getter(z)
        finally:
            # Try to be hygenic.  This is an interpreted language, but do what we can...
            z.close()
            del z
            # Strings in Python are typically immutable; attempts to
            # modify the string will likely just make more copies, so just
            # tell the interpreter to get rid of the main copy asap.
            del plainsrc
    def getZipNames(self):
        return self._getFromZipFile(lambda z: z.namelist())
    def getZipDirectory(self):
        return self._getFromZipFile(lambda z: z.infolist())
    def withZipElementSource(self, elementname, onSrcFunc):
        return self._getFromZipFile(lambda z: onSrcFunc(z.open(elementname, 'rU').read()))
    def find_module(self, fullname, path=None):
        # The fullname indicates which module is to be loaded.  If
        # this import request comes from a module already in the
        # hashedSource, fullname will usually start with the hashRoot
        # (as will path).

        # If this is an unrelated import, path is either None or an
        # array of strings.  If this import is intended for a
        # hashedSource, then path may begin with the hashRoot
        # specification.  When specifically called for the initial
        # import from a Thespian hashedSource, the path will be
        # explicitly passed in as the hashRoot.

        pkgMark = self.hashRoot()
        if path:
            if not hasattr(path, 'startswith') or not path.startswith(pkgMark): return None
            # Both path and fullname may overlap.   For example:
            #   path     = {{hash}}.foo.bar
            #   fullname = {{hash}}.foo.bar.cow.moo
            if not fullname.startswith(path):
                print ('Fullname %s does not start with path %s, aborting'%(fullname, path))
                return None
            skipCnt = len(pkgMark) if fullname.startswith(pkgMark) else 0
            pathname = ospath.join(*tuple(fullname[skipCnt:].split('.')))
        else:
            # No path, but if fullname doesn't start with the
            # hashRoot, then this isn't an import for this sourceHash.
            if not fullname.startswith(pkgMark): return None
            pathname = ospath.join(*tuple(fullname[len(pkgMark):].split('.')))
            if not pathname:
                return HashRootLoader(self)
        for Z in self.getZipDirectory():
            B,E = ospath.splitext(Z.filename)
            if E == '.py':
                if B == pathname:
                    return HashLoader(self)
                if B == pathname + '/__init__':
                    return HashLoader(self, True)
        return None


def loadModuleFromHashSource(sourceHash, sources, modName, modClass):
    if sourceHash not in sources:
        # specified sourceHash does not exist
        logging.getLogger('Thespian').warning('Specified sourceHash %s is not currently loaded',
                                              sourceHash)
        raise InvalidActorSourceHash(sourceHash)

    for metapath in sys.meta_path:
        if getattr(metapath, 'srcHash', None) == sourceHash:
            return _loadModuleFromVerifiedHashSource(metapath, modName, modClass)

    edata = sources[sourceHash]
    f = SourceHashFinder(sourceHash, lambda v: v, edata)
    sys.meta_path.insert(0, f)
    return _loadModuleFromVerifiedHashSource(f, modName, modClass)


def _loadModuleFromVerifiedHashSource(hashFinder, modName, modClass):
    import importlib
    m = importlib.import_module('.'+modName, hashFinder.hashRoot())
    return getattr(m, modClass)
