import unittest
from thespian.test import ActorSystemTestCase
import time
import thespian.test.helpers
from thespian.actors import *
import zipfile
import tempfile
import os


fooSource = '''
from thespian.actors import Actor, requireCapability, PoisonMessage
@requireCapability('Foo Allowed')
class FooActor(Actor):
    def receiveMessage(self, msg, sender):
        if type(msg) == type(""):
            self.send(sender, 'GOT: '+str(msg))
        elif type(msg) == type(1):
            from barn.cow.moo import cow_says  # Import from within this hashSource
            import calendar   # Import a regular module that has not been imported before
            self.send(sender, 'COW: ' + str(cow_says()) + ' on %s'%calendar.weekday(2001,9,11))
        elif type(msg) == type((1,2)):
            if not hasattr(self, 'subA'):
                self.subA = self.createActor('barn.cow.moo.MooActor')
            self.send(self.subA, (msg[1], sender))
        elif isinstance(msg, PoisonMessage) and type(msg.poisonMessage) == type((1,2)):
            self.send(msg.poisonMessage[1], 'FAILED (poisonous)')
'''

mooSource = '''
from thespian.actors import Actor, requireCapability
@requireCapability('Cows Allowed')
class MooActor(Actor):
    def receiveMessage(self, msg, sender):
        if type(msg) == type(""):
            self.send(sender, 'Moo: '+str(msg))
        elif type(msg) == type( (1,2) ):
            self.send(msg[1], 'And MOO: ' + str(msg[0]))

def cow_says():
    return 'Moooo'
'''

dogSource = '''
from thespian.actors import Actor, requireCapability
@requireCapability('Dogs Allowed')
class DogActor(Actor):
    def receiveMessage(self, msg, sender):
        if type(msg) == type(""):
            self.send(sender, 'Woof! '+str(msg))
        elif type(msg) == type( (1,2) ):
            self.send(msg[1], ('Ruff Ruff: ' + str(msg[0]), sender))
'''


class BarActor(Actor):
    def receiveMessage(self, msg, sender):
        if type(msg) == type(""):
            self.send(sender, 'SAW: '+str(msg))


class TestASimpleSystem(ActorSystemTestCase):
    testbase='Simple'
    scope='func'
    actorSystemBase = 'simpleSystemBase'

    def setUp(self):

        self.tmpdir = tempfile.mkdtemp()

        self.foozipFname = os.path.join(self.tmpdir, 'foosrc.zip')
        foozip = zipfile.ZipFile(self.foozipFname, 'w')
        foozip.writestr('foo.py', fooSource)
        foozip.writestr('barn/__init__.py', '')
        foozip.writestr('barn/cow/__init__.py', '')
        foozip.writestr('barn/cow/moo.py', mooSource)
        foozip.close()

        self.foozipEncFile = self._encryptROT13Zipfile(self.foozipFname)

        self.dogzipFname = os.path.join(self.tmpdir, 'dogsrc.zip')
        dogzip = zipfile.ZipFile(self.dogzipFname, 'w')
        dogzip.writestr('dog.py', dogSource)
        dogzip.close()

        self.dogzipEncFile = self._encryptROT13Zipfile(self.dogzipFname)

        self.systems = {}
        # Only define base capabilities, not extended capabilities
        self.capabilities = { 'One': { 'Admin Port': 12321,
                                       'Foo Allowed': True,
                                       'Cows Allowed': True,
                                       'Dogs Allowed': True,
                                   },
                          }
        for each in ['One']:  # 'One' must be first
            self.systems[each] = ActorSystem(self.actorSystemBase, self.capabilities[each],
                                             logDefs = ActorSystemTestCase.getDefaultTestLogging())
            time.sleep(0.1)  # Wait for each ActorSystem to start

    def tearDown(self):
        for each in self.capabilities:
            self.systems[each].shutdown()
        import shutil
        if os.path.exists(self.tmpdir): shutil.rmtree(self.tmpdir)
        time.sleep(0.1)  # Wait for ActorSystems to stop

    def test00_systemsRunnable(self):
        pass

    def test01_verifyFooActorNotAvailableByName(self):
        self.assertRaises(ImportError, self.systems['One'].createActor, 'foo.FooActor')
        bar = self.systems['One'].createActor('thespian.test.testLoadSource.BarActor')
        self.assertEqual('SAW: hello', self.systems['One'].ask(bar, 'hello', 1))

    def test01_verifyFooActorNotAvailableWithBogusSourceHash(self):
        self.assertRaises(InvalidActorSourceHash,
                          self.systems['One'].createActor,
                          'foo.FooActor', sourceHash = 'this is bogus')

    def test01_verifyloadSourceHandlesBadFilename(self):
        self.assertRaises(IOError,
                          self.systems['One'].loadActorSource, 'bad file name here')


    def _encryptROT13Zipfile(self, zipFname):
        zFile = open(zipFname, 'rb')
        zData = zFile.read()
        zFile.close()
        rot13 = lambda b: chr((ord(b) + 13) % 256)
        z = 'ROT13___' + ''.join(map(rot13, zData))
        efName = zipFname + '.enc'
        zEFile = open(efName, 'wb')
        zEFile.write(z)
        zEFile.close()
        return efName

    def _loadFooSource(self):
        srchash = self.systems['One'].loadActorSource(self.foozipFname)
        self.assertIsNotNone(srchash)
        return srchash

    def test02_verifyMainActorAvailableWhenLoaded(self):
        srchash = self._loadFooSource()
        foo = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash)
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))

    def test02_verifySubActorAvailableWhenLoaded(self):
        srchash = self._loadFooSource()
        foo = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash)
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))
        self.assertEqual('And MOO: great', self.systems['One'].ask(foo, ('discard', 'great')))

    def test02_verifySubModuleAvailableWhenLoaded(self):
        srchash = self._loadFooSource()
        cow = self.systems['One'].createActor('barn.cow.moo.MooActor', sourceHash=srchash)
        self.assertEqual('Moo: got milk', self.systems['One'].ask(cow, 'got milk', 1))

    def test02_verifyHashSourceAvailablePostLoadFromMembers(self):
        srchash = self._loadFooSource()
        foo = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash)
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))
        # Cause a load-on-demand of a new module (in the loaded
        # sources) from the loaded sources themselves and ensure it's still available.
        self.assertEqual('COW: Moooo on 1', self.systems['One'].ask(foo, 1))

    def test02_verifyHashSourceNotInGlobalNamespace(self):
        srchash = self._loadFooSource()
        foo = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash)
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))
        # Verify loaded source is not accessible globally
        try:
            from bar.cow.moo import cow_says
            self.assertTrue(False)  # should never get here
        except ImportError:
            self.assertTrue(True)   # want this
        except Exception:
            self.assertTrue(False)  # but not these


    # Note: Perform these after the successful load tests (test02_) to
    # ensure that the modules loaded by those tests are no longer
    # available in the namespace to cause these tests (test03_) to fail.
    def test03_verifyFooActorNotAvailableWithoutModuleQualifiersOrHash(self):
        srchash = self._loadFooSource()
        self.assertRaises(InvalidActorSpecification,
                          self.systems['One'].createActor, 'FooActor')

    def test03_verifyFooActorNotAvailableWithoutCorrectHash(self):
        # No sourceHash specified, so the module foo is searched for
        # in the standard search path.
        srchash = self._loadFooSource()
        self.assertRaises(ImportError, self.systems['One'].createActor, 'foo.FooActor')

    def test03_verifyFooActorNotAvailableWithoutModuleQualifiers(self):
        srchash = self._loadFooSource()
        # The FooActor is not available without module qualifiers even if proper hash is specified
        self.assertRaises(InvalidActorSpecification,
                          self.systems['One'].createActor, 'FooActor',
                          sourceHash = srchash)

    def test04_verifyReloadOfChangedModuleAllowsBothToExistSimultaneously(self):
        srchash = self._loadFooSource()
        foo = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash)
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))
        self.assertEqual('And MOO: great', self.systems['One'].ask(foo, ('discard', 'great')))

        # Update the foo sources
        foo2zipFname = os.path.join(self.tmpdir, 'foo2src.zip')
        foozip = zipfile.ZipFile(foo2zipFname, 'w')
        foozip.writestr('foo.py', fooSource.replace('GOT:', 'TOG:'))
        foozip.writestr('barn/__init__.py', '')
        foozip.writestr('barn/cow/__init__.py', '')
        foozip.writestr('barn/cow/moo.py', mooSource.replace('And MOO:', '& MOO:'))
        foozip.close()

        # Load the updated foo sources... next to the original
        srchash2 = self.systems['One'].loadActorSource(foo2zipFname)
        self.assertIsNotNone(srchash2)
        self.assertNotEqual(srchash, srchash2)

        foo2 = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash2)
        self.assertEqual('TOG: good one', self.systems['One'].ask(foo2, 'good one', 1))
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))
        self.assertEqual('& MOO: great', self.systems['One'].ask(foo2, ('discard', 'great'), 1))
        self.assertEqual('And MOO: great', self.systems['One'].ask(foo, ('discard', 'great'), 1))

        self.assertEqual('COW: Moooo on 1', self.systems['One'].ask(foo, 1, 1))
        self.assertEqual('COW: Moooo on 1', self.systems['One'].ask(foo2, 1, 1))

    def test04_verifyMultipleSeparateModulesLoaded(self):
        srchash = self._loadFooSource()
        srchash2 = self.systems['One'].loadActorSource(self.dogzipFname)
        self.assertIsNotNone(srchash2)
        foo = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash)
        dog = self.systems['One'].createActor('dog.DogActor', sourceHash=srchash2)
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))
        self.assertEqual('And MOO: great', self.systems['One'].ask(foo, ('discard', 'great')))
        self.assertEqual('Woof! bark', self.systems['One'].ask(dog, 'bark', 1))

    def test04_verifyMultipleSeparateModulesRequireCorrectSourceHashOnCreate(self):
        srchash = self._loadFooSource()
        srchash2 = self.systems['One'].loadActorSource(self.dogzipFname)
        self.assertIsNotNone(srchash2)
        foo = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash)
        self.assertRaises(ImportError, self.systems['One'].createActor,
                          'dog.DogActor', sourceHash=srchash)  # wrong source hash
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))

    def test05_verifyUnloadOfHashedSourcePreventsActorCreation(self):
        srchash = self._loadFooSource()
        foo = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash)
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))
        self.assertEqual('And MOO: great', self.systems['One'].ask(foo, ('discard', 'great'), 1))
        # Unload fooSource
        self.systems['One'].unloadActorSource(srchash)
        # Test cannot create actors anymore
        self.assertRaises(InvalidActorSourceHash,
                          self.systems['One'].createActor,
                          'foo.FooActor', sourceHash = srchash)

    def test05_verifyUnloadOfHashedSourceDoesNotKillActiveActors(self):
        srchash = self._loadFooSource()
        foo = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash)
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))
        self.assertEqual('And MOO: great', self.systems['One'].ask(foo, ('discard', 'great'), 1))
        # Unload fooSource
        self.systems['One'].unloadActorSource(srchash)
        # Test foo actors still exist
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))
        self.assertEqual('And MOO: great', self.systems['One'].ask(foo, ('discard', 'great')))


    # Unloading an actor source has indeterminate effects on running
    # Actors.  For actors that are part of the current process
    # (e.g. simpleSystemBase, multi-threaded system bases) the unload
    # will probably make the source unavailable for running actors as
    # well.  For actors in a multi-process configuration, the unload
    # at the local/admin point will not likely affect running actors
    # (this would have to be implemented by propagating the unload to
    # all other actors, which is of questionable benefit compared to
    # the overhead and timing issue).  For this reason, the following
    # test is NOT performed.

    # def test05_verifyUnloadOfHashedSourceDoesNotAllowNewSubActorsToBeCreated(self):
    #     srchash = self._loadFooSource()
    #     foo = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash)
    #     self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))
    #     # Unload fooSource
    #     self.systems['One'].unloadActorSource(srchash)
    #     # Verify cannot create sub-actor from still-existing actor from unloaded source

    #     # First, show that an attempt to create an actor that was in
    #     # the removed source gives back an actorAddress to the
    #     # creating Actor, but the actor is never created and thus the
    #     # ask will timeout.
    #     self.assertIsNone(self.systems['One'].ask(foo, ('discard', 'great'), 0.25))

    #     # Now show that an attempt to get a running actor to import a
    #     # module that has been unloaded will cause that Actor to fail
    #     # on the import, resulting in a PoisonMessage indication.
    #     r = self.systems['One'].ask(foo, 1, 0.25)
    #     self.assertIsInstance(r, PoisonMessage)
    #     self.assertEqual(r.poisonMessage, 1)

    def test05_verifyMultipleSeparateModulesCanUseOtherAfterFirstUnloaded(self):
        srchash = self._loadFooSource()
        srchash2 = self.systems['One'].loadActorSource(self.dogzipFname)
        self.assertIsNotNone(srchash2)
        self.systems['One'].unloadActorSource(srchash)
        self.assertRaises(InvalidActorSourceHash,
                          self.systems['One'].createActor,
                          'foo.FooActor', sourceHash=srchash)
        dog = self.systems['One'].createActor('dog.DogActor', sourceHash=srchash2)
        self.assertEqual('Woof! bark', self.systems['One'].ask(dog, 'bark', 1))

    def test04_verifyReloadOfChangedModuleAndUnloadOfOriginal(self):
        srchash = self._loadFooSource()
        foo = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash)
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))
        self.assertEqual('And MOO: great', self.systems['One'].ask(foo, ('discard', 'great')))

        # Update the foo sources
        foo2zipFname = os.path.join(self.tmpdir, 'foo2src.zip')
        foozip = zipfile.ZipFile(foo2zipFname, 'w')
        foozip.writestr('foo.py', fooSource.replace('GOT:', 'TOG:'))
        foozip.writestr('barn/__init__.py', '')
        foozip.writestr('barn/cow/__init__.py', '')
        foozip.writestr('barn/cow/moo.py', mooSource.replace('And MOO:', '& MOO:'))
        foozip.close()

        # Load the updated foo sources... next to the original
        srchash2 = self.systems['One'].loadActorSource(foo2zipFname)
        self.assertIsNotNone(srchash2)
        self.assertNotEqual(srchash, srchash2)

        self.systems['One'].unloadActorSource(srchash)
        self.systems['One'].tell(foo, ActorExitRequest())

        foo2 = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash2)
        self.assertRaises(InvalidActorSourceHash,
                          self.systems['One'].createActor,
                          'foo.FooActor', sourceHash=srchash)
        self.assertEqual('TOG: good one', self.systems['One'].ask(foo2, 'good one', 1))
        self.assertEqual('& MOO: great', self.systems['One'].ask(foo2, ('discard', 'great'), 1))

        self.assertEqual('COW: Moooo on 1', self.systems['One'].ask(foo2, 1, 1))

    def test06_sourceAuthorityCanRegister(self):
        auth = self.systems['One'].createActor(rot13Authority)
        enabled = self.systems['One'].ask(auth, 'Enable', 1)
        self.assertEqual(enabled, 'Enabled')

    def test07_sourceAuthorityRejectsInvalidSource(self):
        auth = self.systems['One'].createActor(rot13Authority)
        enabled = self.systems['One'].ask(auth, 'Enable', 1)
        self.assertEqual(enabled, 'Enabled')
        srchash = self._loadFooSource()
        self.assertIsNotNone(srchash)
        time.sleep(0.25)  # allow time for load to consult Source Authority
        self.assertRaises(InvalidActorSourceHash,
                          self.systems['One'].createActor,
                          'foo.FooActor', sourceHash = srchash)

    def test07_sourceAuthorityAcceptsValidSource(self):
        auth = self.systems['One'].createActor(rot13Authority)
        enabled = self.systems['One'].ask(auth, 'Enable', 1)
        self.assertEqual(enabled, 'Enabled')
        srchash = self.systems['One'].loadActorSource(self.foozipEncFile)
        self.assertIsNotNone(srchash)
        time.sleep(0.25)  # allow time for load to consult Source Authority
        foo = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash)
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))
        self.assertEqual('And MOO: great', self.systems['One'].ask(foo, ('discard', 'great')))

    def test07_sourceAuthorityAcceptsValidSourceAfterBadSource(self):
        auth = self.systems['One'].createActor(rot13Authority)
        enabled = self.systems['One'].ask(auth, 'Enable', 1)
        self.assertEqual(enabled, 'Enabled')
        failhash = self.systems['One'].loadActorSource(self.dogzipFname)
        srchash = self.systems['One'].loadActorSource(self.foozipEncFile)
        self.assertIsNotNone(srchash)
        time.sleep(0.25)  # allow time for load to consult Source Authority
        foo = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash)
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))
        self.assertEqual('And MOO: great', self.systems['One'].ask(foo, ('discard', 'great')))
        self.assertRaises(InvalidActorSourceHash,
                          self.systems['One'].createActor,
                          'dog.DogActor', sourceHash = failhash)

    def test07_sourceAuthorityAcceptsMultipleValidSources(self):
        auth = self.systems['One'].createActor(rot13Authority)
        enabled = self.systems['One'].ask(auth, 'Enable', 1)
        self.assertEqual(enabled, 'Enabled')
        srchash = self.systems['One'].loadActorSource(self.foozipEncFile)
        self.assertIsNotNone(srchash)
        srchash2 = self.systems['One'].loadActorSource(self.dogzipEncFile)
        self.assertIsNotNone(srchash2)
        time.sleep(0.25)  # allow time for loads to consult Source Authority
        foo = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash)
        dog = self.systems['One'].createActor('dog.DogActor', sourceHash=srchash2)
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))
        self.assertEqual('And MOO: great', self.systems['One'].ask(foo, ('discard', 'great')))
        self.assertEqual('Woof! bark', self.systems['One'].ask(dog, 'bark', 1))

    def test07_multipleValidSourcesCanCommunicate(self):
        auth = self.systems['One'].createActor(rot13Authority)
        enabled = self.systems['One'].ask(auth, 'Enable', 1)
        self.assertEqual(enabled, 'Enabled')
        srchash = self.systems['One'].loadActorSource(self.foozipEncFile)
        self.assertIsNotNone(srchash)
        srchash2 = self.systems['One'].loadActorSource(self.dogzipEncFile)
        self.assertIsNotNone(srchash2)
        time.sleep(0.25)  # allow time for loads to consult Source Authority
        cow = self.systems['One'].createActor('barn.cow.moo.MooActor', sourceHash=srchash)
        dog = self.systems['One'].createActor('dog.DogActor', sourceHash=srchash2)
        self.assertEqual('Moo: good one', self.systems['One'].ask(cow, 'good one', 1))
        self.assertEqual('Woof! good boy', self.systems['One'].ask(dog, 'good boy', 1))
        self.assertEqual('And MOO: Ruff Ruff: hungry',
                         self.systems['One'].ask(dog, ('hungry', cow), 1))

    def test07_sourceAuthorityAcceptsValidSourceResultIsCorrupted(self):
        auth = self.systems['One'].createActor(rot13CorruptAuthority)
        enabled = self.systems['One'].ask(auth, 'Enable', 1)
        self.assertEqual(enabled, 'Enabled')
        srchash = self.systems['One'].loadActorSource(self.foozipEncFile)
        self.assertIsNotNone(srchash)
        time.sleep(0.25)  # allow time for load to consult Source Authority
        self.assertRaises(InvalidActorSourceHash,
                          self.systems['One'].createActor,
                          'foo.FooActor', sourceHash = srchash)

    def test07_sourceAuthorityExceptions(self):
        auth = self.systems['One'].createActor(rot13FailAuthority)
        enabled = self.systems['One'].ask(auth, 'Enable', 1)
        self.assertEqual(enabled, 'Enabled')
        srchash = self.systems['One'].loadActorSource(self.foozipEncFile)
        self.assertIsNotNone(srchash)
        time.sleep(0.25)  # allow time for load to consult Source Authority
        self.assertRaises(InvalidActorSourceHash,
                          self.systems['One'].createActor,
                          'foo.FooActor', sourceHash = srchash)

class rot13Authority(Actor):
    def receiveMessage(self, msg, sender):
        if msg == 'Enable':
            self.registerSourceAuthority()
            self.send(sender, 'Enabled')
        elif msg == 'CorruptNext':
            self.corruptNext = True
            self.send(sender, 'OK')
        elif isinstance(msg, ValidateSource):
            if msg.sourceData[:8] == 'ROT13___':
                unrot13 = lambda b: chr((ord(b) + 256 - 13) % 256)
                clear = ''.join(map(unrot13, msg.sourceData[8:]))
                if getattr(self, 'corruptNext', False):
                    corruption = 'corrupted'
                    for x in range(5, len(clear), len(corruption) + 100):
                        clear = clear[:x] + corruption + clear[x+len(corruption):]
                self.send(sender, ValidatedSource(msg.sourceHash, clear))

class rot13CorruptAuthority(Actor):
    def receiveMessage(self, msg, sender):
        if msg == 'Enable':
            self.registerSourceAuthority()
            self.send(sender, 'Enabled')
        elif isinstance(msg, ValidateSource):
            if msg.sourceData[:8] == 'ROT13___':
                unrot13 = lambda b: chr((ord(b) + 256 - 13) % 256)
                clear = ''.join(map(unrot13, msg.sourceData[8:]))
                corruption = 'corrupted'
                for x in range(5, len(clear), len(corruption) + 100):
                    clear = clear[:x] + corruption + clear[x+len(corruption):]
                self.send(sender, ValidatedSource(msg.sourceHash, clear))

class rot13FailAuthority(Actor):
    def receiveMessage(self, msg, sender):
        if msg == 'Enable':
            self.registerSourceAuthority()
            self.send(sender, 'Enabled')
        elif isinstance(msg, ValidateSource):
            if msg.sourceData[:8] == 'ROT13___':
                raise ValueError('Oh no, I must go')


class TestMultiProcTCPSystem(TestASimpleSystem):
    testbase='MultiprocTCP'
    scope='func'
    actorSystemBase = 'multiprocTCPBase'


class TestMultiProcUDPSystem(TestASimpleSystem):
    testbase='MultiprocUDP'
    scope='func'
    actorSystemBase = 'multiprocUDPBase'

class TestMultiProcQueueSystem(TestASimpleSystem):
    testbase='MultiprocQueue'
    scope='func'
    actorSystemBase = 'multiprocQueueBase'



class TestMultipleMultiProcTCPSystem(ActorSystemTestCase):
    testbase='MultiprocTCP'
    scope='func'
    actorSystemBase = 'multiprocTCPBase'

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()

        self.foozipFname = os.path.join(self.tmpdir, 'foosrc.zip')
        foozip = zipfile.ZipFile(self.foozipFname, 'w')
        foozip.writestr('foo.py', fooSource)
        foozip.writestr('barn/__init__.py', '')
        foozip.writestr('barn/cow/__init__.py', '')
        foozip.writestr('barn/cow/moo.py', mooSource)
        foozip.close()

        self.foozipEncFile = self._encryptROT13Zipfile(self.foozipFname)

        self.dogzipFname = os.path.join(self.tmpdir, 'dogsrc.zip')
        dogzip = zipfile.ZipFile(self.dogzipFname, 'w')
        dogzip.writestr('dog.py', dogSource)
        dogzip.close()

        self.dogzipEncFile = self._encryptROT13Zipfile(self.dogzipFname)

        self.systems = {}
        # Only define base capabilities, not extended capabilities
        self.capabilities = { 'One': { 'Admin Port': 12321, },
                              'Two': { 'Admin Port': 12322,
                                       'Convention Address.IPv4': ('', 12321), },
                          }
        for each in ['One', 'Two']:  # 'One' must be first
            self.systems[each] = ActorSystem(self.actorSystemBase, self.capabilities[each],
                                             logDefs = ActorSystemTestCase.getDefaultTestLogging())
        time.sleep(0.1)  # Wait for Actor Systems to start

    def tearDown(self):
        for each in self.capabilities:
            self.systems[each].shutdown()
        import shutil
        if os.path.exists(self.tmpdir): shutil.rmtree(self.tmpdir)

    def _encryptROT13Zipfile(self, zipFname):
        zFile = open(zipFname, 'rb')
        zData = zFile.read()
        zFile.close()
        rot13 = lambda b: chr((ord(b) + 13) % 256)
        z = 'ROT13___' + ''.join(map(rot13, zData))
        efName = zipFname + '.enc'
        zEFile = open(efName, 'wb')
        zEFile.write(z)
        zEFile.close()
        return efName

    def test00_systemsRunnable(self):
        pass

    def test09_properErrorIfActorCapabilitiesNotSatisfied(self):
        # Specify a Source Authority and load the foo sources
        auth = self.systems['One'].createActor(rot13Authority)
        enabled = self.systems['One'].ask(auth, 'Enable', 1)
        self.assertEqual(enabled, 'Enabled')
        srchash = self.systems['One'].loadActorSource(self.foozipEncFile)
        self.assertIsNotNone(srchash)
        # Establish capabilities that allow Foo and Moo actors (in different systems)
        time.sleep(0.25)  # allow time for load to consult Source Authority
        self.assertRaises(NoCompatibleSystemForActor,
                          self.systems['One'].createActor,
                          'foo.FooActor', sourceHash=srchash)


    def test08_multiSystemSharesLoadedSourcesByDefault(self):
        # Specify a Source Authority and load the foo sources
        auth = self.systems['One'].createActor(rot13Authority)
        enabled = self.systems['One'].ask(auth, 'Enable', 1)
        self.assertEqual(enabled, 'Enabled')
        srchash = self.systems['One'].loadActorSource(self.foozipEncFile)
        self.assertIsNotNone(srchash)
        self.systems['One'].updateCapability('Foo Allowed', True)
        self.systems['Two'].updateCapability('Cows Allowed', True)
        # Establish capabilities that allow Foo and Moo actors (in different systems)
        time.sleep(0.25)  # allow time for load to consult Source Authority

        # Verify that FooActor can be created (locally) and it can
        # create MooActor remotely, where the remote system will
        # obtain the sources from the local system as needed.
        foo = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash)
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))
        self.assertEqual('And MOO: great',
                         self.systems['One'].ask(foo, ('discard', 'great'), 1))

    def test08_loadableInRemoteSystemMatchingCapabilities(self):
        # Specify a Source Authority and load the foo sources
        auth = self.systems['One'].createActor(rot13Authority)
        enabled = self.systems['One'].ask(auth, 'Enable', 1)
        self.assertEqual(enabled, 'Enabled')
        srchash = self.systems['One'].loadActorSource(self.foozipEncFile)
        self.assertIsNotNone(srchash)
        self.systems['Two'].updateCapability('Foo Allowed', True)
        self.systems['One'].updateCapability('Cows Allowed', True)
        # Establish capabilities that allow Foo and Moo actors (in different systems)
        time.sleep(0.25)  # allow time for load to consult Source Authority

        # Verify that FooActor can be created (locally) and it can
        # create MooActor remotely, where the remote system will
        # obtain the sources from the local system as needed.
        foo = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash)
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))
        self.assertEqual('And MOO: great',
                         self.systems['One'].ask(foo, ('discard', 'great'), 1))

    def test08_multiSystemSharesLoadedSourcesIfExplicitlyAllowed(self):
        # Specify a Source Authority and load the foo sources
        auth = self.systems['One'].createActor(rot13Authority)
        enabled = self.systems['One'].ask(auth, 'Enable', 1)
        self.assertEqual(enabled, 'Enabled')
        srchash = self.systems['One'].loadActorSource(self.foozipEncFile)
        self.assertIsNotNone(srchash)
        self.systems['One'].updateCapability('Foo Allowed', True)
        self.systems['Two'].updateCapability('Cows Allowed', True)
        self.systems['Two'].updateCapability('AllowRemoteActorSources', 'yes')
        # Establish capabilities that allow Foo and Moo actors (in different systems)
        time.sleep(0.25)  # allow time for load to consult Source Authority

        # Verify that FooActor can be created (locally) and it can
        # create MooActor remotely, where the remote system will
        # obtain the sources from the local system as needed.
        foo = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash)
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))
        self.assertEqual('And MOO: great',
                         self.systems['One'].ask(foo, ('discard', 'great'), 1))

    def test08_multiSystemLoadedSourcesNotSharedIfExplicitlyDisallowed(self):
        # Specify a Source Authority and load the foo sources
        auth = self.systems['One'].createActor(rot13Authority)
        enabled = self.systems['One'].ask(auth, 'Enable', 1)
        self.assertEqual(enabled, 'Enabled')
        srchash = self.systems['One'].loadActorSource(self.foozipEncFile)
        self.assertIsNotNone(srchash)
        self.systems['One'].updateCapability('Foo Allowed', True)
        self.systems['Two'].updateCapability('Cows Allowed', True)
        self.systems['Two'].updateCapability('AllowRemoteActorSources', 'no')
        # Establish capabilities that allow Foo and Moo actors (in different systems)
        time.sleep(0.25)  # allow time for load to consult Source Authority

        # Verify that FooActor can be created (locally) and it can
        # create MooActor remotely, where the remote system will
        # obtain the sources from the local system as needed.
        foo = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash)
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))
        self.assertEqual('FAILED (poisonous)',
                         self.systems['One'].ask(foo, ('discard', 'great'), 1))

    def test08_multiSystemLoadedSourcesNotSharedIfSharingUnrecognized(self):
        # Specify a Source Authority and load the foo sources
        auth = self.systems['One'].createActor(rot13Authority)
        enabled = self.systems['One'].ask(auth, 'Enable', 1)
        self.assertEqual(enabled, 'Enabled')
        srchash = self.systems['One'].loadActorSource(self.foozipEncFile)
        self.assertIsNotNone(srchash)
        self.systems['One'].updateCapability('Foo Allowed', True)
        self.systems['Two'].updateCapability('Cows Allowed', True)
        self.systems['Two'].updateCapability('AllowRemoteActorSources', 'whatever!')
        # Establish capabilities that allow Foo and Moo actors (in different systems)
        time.sleep(0.25)  # allow time for load to consult Source Authority

        # Verify that FooActor can be created (locally) and it can
        # create MooActor remotely, where the remote system will
        # obtain the sources from the local system as needed.
        foo = self.systems['One'].createActor('foo.FooActor', sourceHash=srchash)
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))
        self.assertEqual('FAILED (poisonous)',
                         self.systems['One'].ask(foo, ('discard', 'great'), 1))

    def test08_loadableInRemoteSystemOnlyIfSourceComesFromConventionLeader(self):
        auth = self.systems['Two'].createActor(rot13Authority)
        enabled = self.systems['Two'].ask(auth, 'Enable', 1)
        self.assertEqual(enabled, 'Enabled')
        srchash = self.systems['Two'].loadActorSource(self.foozipEncFile)
        self.assertIsNotNone(srchash)
        time.sleep(0.1)

        self.systems['Two'].updateCapability('Foo Allowed', True)
        self.systems['One'].updateCapability('Cows Allowed', True)
        self.systems['One'].updateCapability('AllowRemoteActorSources', 'LeaderOnly')
        self.systems['Two'].updateCapability('AllowRemoteActorSources', 'LeaderOnly')
        # Establish capabilities that allow Foo and Moo actors (in different systems)
        time.sleep(0.25)  # allow time for load to consult Source Authority

        self.assertRaises(InvalidActorSourceHash,
                          self.systems['One'].createActor,
                          'foo.FooActor', sourceHash=srchash)

        self.systems['One'].updateCapability('AllowRemoteActorSources', 'yes')
        time.sleep(0.08)
        foo = self.systems['Two'].createActor('foo.FooActor', sourceHash=srchash)
        self.assertEqual('GOT: good one', self.systems['One'].ask(foo, 'good one', 1))
        self.assertEqual('And MOO: great',
                         self.systems['One'].ask(foo, ('discard', 'great'), 1))

class TestMultipleMultiProcUDPSystem(TestMultipleMultiProcTCPSystem):
    testbase='MultiprocUDP'
    scope='func'
    actorSystemBase = 'multiprocUDPBase'

# Support multiple simpleactorsystems?  shared runSends?
# sourceData held in mildly-encrypted form instead of plaincode

# KWQ: should AllowRemoteActorSources only be settable on initial load and not subsequently changeable to prevent unlocking injections?  On startup, provide a key; all ActorSystem api elements must supply the same key to authenticate themselves?

# Also need to authenticate messages exchanged between valid Actor agents to prevent injection around the ActorSytem?

# Have security levels:  0 - plaintext, 1 - random key embedded privately in ACtorAccresses, 2 - encrypted messages with a key

